from __future__ import annotations

import sys
from typing import Any

if sys.version_info < (3, 10, 0):
    import importlib_resources as resources
else:
    from importlib import resources  # type: ignore[no-redef]

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from joblib import dump
from mqt.bench.utils import calc_supermarq_features
from mqt.predictor import ml, reward, rl
from qiskit import QuantumCircuit

if TYPE_CHECKING:
    from numpy._typing import NDArray
    from sklearn.ensemble import RandomForestClassifier


def qcompile(
    qc: QuantumCircuit, figure_of_merit: reward.figure_of_merit = "expected_fidelity"
) -> tuple[QuantumCircuit, list[str], str] | bool:
    """Compiles a given quantum circuit to a device with the highest predicted figure of merit.

    Args:
        qc (QuantumCircuit): The quantum circuit to be compiled.
        figure_of_merit (reward.reward_functions, optional): The figure of merit to be used for compilation. Defaults to "expected_fidelity".

    Returns:
        tuple[QuantumCircuit, list[str], str] | bool: Returns a tuple containing the compiled quantum circuit, the compilation information and the name of the device used for compilation. If compilation fails, False is returned.
    """

    device_name = predict_device_for_figure_of_merit(qc, figure_of_merit)
    assert device_name is not None
    res = rl.qcompile(qc, figure_of_merit=figure_of_merit, device_name=device_name)
    return *res, device_name


def predict_device_for_figure_of_merit(
    qc: QuantumCircuit, figure_of_merit: reward.figure_of_merit = "expected_fidelity"
) -> str:
    """Returns the name of the device with the highest predicted figure of merit that is suitable for the given quantum circuit.

    Args:
        qc (QuantumCircuit): The quantum circuit to be compiled.
        figure_of_merit (reward.reward_functions, optional): The figure of merit to be used for compilation. Defaults to "expected_fidelity".

    Returns:
        str : The name of the device with the highest predicted figure of merit that is suitable for the given quantum circuit.
    """

    ml_predictor = ml.Predictor()
    predicted_device_index_probs = ml_predictor.predict_probs(qc, figure_of_merit)
    assert ml_predictor.clf is not None
    classes = ml_predictor.clf.classes_  # type: ignore[unreachable]
    predicted_device_index = classes[np.argsort(predicted_device_index_probs)[::-1]]
    devices = rl.helper.get_devices()

    for index in predicted_device_index:
        if devices[index]["max_qubits"] >= qc.num_qubits:
            return devices[index]["name"]
    msg = "No suitable device found."
    raise ValueError(msg)


def get_path_training_data() -> Path:
    """Returns the path to the training data folder."""
    return Path(str(resources.files("mqt.predictor"))) / "ml" / "training_data"


def get_path_results(ghz_results: bool = False) -> Path:
    """Returns the path to the results file."""
    if ghz_results:
        return get_path_training_data() / "trained_model" / "res_GHZ.csv"
    return get_path_training_data() / "trained_model" / "res.csv"


def get_path_trained_model(figure_of_merit: str, return_non_zero_indices: bool = False) -> Path:
    """Returns the path to the trained model folder resulting from the machine learning training."""
    if return_non_zero_indices:
        return get_path_training_data() / "trained_model" / ("non_zero_indices_" + figure_of_merit + ".npy")
    return get_path_training_data() / "trained_model" / ("trained_clf_" + figure_of_merit + ".joblib")


def get_path_training_circuits() -> Path:
    """Returns the path to the training circuits folder."""
    return get_path_training_data() / "training_circuits"


def get_path_training_circuits_compiled() -> Path:
    """Returns the path to the compiled training circuits folder."""
    return get_path_training_data() / "training_circuits_compiled"


def get_index_to_device_LUT() -> dict[int, str]:
    """Returns a look-up table (LUT) that maps the index of a device to its name."""
    devices = rl.helper.get_devices()
    return {i: device["name"] for i, device in enumerate(devices)}


def get_openqasm_gates() -> list[str]:
    """Returns a list of all quantum gates within the openQASM 2.0 standard header."""
    # according to https://github.com/Qiskit/qiskit-terra/blob/main/qiskit/qasm/libs/qelib1.inc
    return [
        "u3",
        "u2",
        "u1",
        "cx",
        "id",
        "u0",
        "u",
        "p",
        "x",
        "y",
        "z",
        "h",
        "s",
        "sdg",
        "t",
        "tdg",
        "rx",
        "ry",
        "rz",
        "sx",
        "sxdg",
        "cz",
        "cy",
        "swap",
        "ch",
        "ccx",
        "cswap",
        "crx",
        "cry",
        "crz",
        "cu1",
        "cp",
        "cu3",
        "csx",
        "cu",
        "rxx",
        "rzz",
        "rccx",
        "rc3x",
        "c3x",
        "c3sqrtx",
        "c4x",
    ]


def dict_to_featurevector(gate_dict: dict[str, int]) -> dict[str, int]:
    """Calculates and returns the feature vector of a given quantum circuit gate dictionary."""
    res_dct = dict.fromkeys(get_openqasm_gates(), 0)
    for key, val in dict(gate_dict).items():
        if key in res_dct:
            res_dct[key] = val

    return res_dct


PATH_LENGTH = 260


def create_feature_dict(qc: str | QuantumCircuit) -> dict[str, Any]:
    """Creates and returns a feature dictionary for a given quantum circuit.

    Args:
        qc (str | QuantumCircuit): The quantum circuit to be compiled.

    Returns:
        dict[str, Any]: The feature dictionary of the given quantum circuit.
    """
    if not isinstance(qc, QuantumCircuit):
        if len(qc) < PATH_LENGTH and Path(qc).exists():
            qc = QuantumCircuit.from_qasm_file(qc)
        elif "OPENQASM" in qc:
            qc = QuantumCircuit.from_qasm_str(qc)
        else:
            error_msg = "Invalid input for 'qc' parameter."
            raise ValueError(error_msg) from None

    ops_list = qc.count_ops()
    ops_list_dict = dict_to_featurevector(ops_list)

    feature_dict = {}
    for key in ops_list_dict:
        feature_dict[key] = float(ops_list_dict[key])

    feature_dict["num_qubits"] = float(qc.num_qubits)
    feature_dict["depth"] = float(qc.depth())

    supermarq_features = calc_supermarq_features(qc)
    feature_dict["program_communication"] = supermarq_features.program_communication
    feature_dict["critical_depth"] = supermarq_features.critical_depth
    feature_dict["entanglement_ratio"] = supermarq_features.entanglement_ratio
    feature_dict["parallelism"] = supermarq_features.parallelism
    feature_dict["liveness"] = supermarq_features.liveness
    return feature_dict


def save_classifier(clf: RandomForestClassifier, figure_of_merit: reward.figure_of_merit = "expected_fidelity") -> None:
    """Saves the given classifier to the trained model folder.

    Args:
        clf (RandomForestClassifier): The classifier to be saved.
        figure_of_merit (reward.reward_functions, optional): The figure of merit to be used for compilation. Defaults to "expected_fidelity".
    """
    dump(clf, str(get_path_trained_model(figure_of_merit)))


def save_training_data(
    training_data: list[NDArray[np.float_]],
    names_list: list[str],
    scores_list: list[NDArray[np.float_]],
    figure_of_merit: reward.figure_of_merit,
) -> None:
    """Saves the given training data to the training data folder.

    Args:
        res (tuple[list[Any], list[Any], list[Any]]): The training data, the names list and the scores list to be saved.
        figure_of_merit (reward.reward_functions, optional): The figure of merit to be used for compilation. Defaults to "expected_fidelity".
    """

    with resources.as_file(get_path_training_data() / "training_data_aggregated") as path:
        data = np.asarray(training_data, dtype=object)
        np.save(str(path / ("training_data_" + figure_of_merit + ".npy")), data)
        data = np.asarray(names_list)
        np.save(str(path / ("names_list_" + figure_of_merit + ".npy")), data)
        data = np.asarray(scores_list)
        np.save(str(path / ("scores_list_" + figure_of_merit + ".npy")), data)


def load_training_data(
    figure_of_merit: reward.figure_of_merit = "expected_fidelity",
) -> tuple[NDArray[np.float_], list[str], list[NDArray[np.float_]]]:
    """Loads and returns the training data from the training data folder.

    Args:
        figure_of_merit (reward.reward_functions, optional): The figure of merit to be used for compilation. Defaults to "expected_fidelity".

    Returns:
       tuple[NDArray[np.float_], list[str], list[NDArray[np.float_]]]: The training data, the names list and the scores list.
    """
    with resources.as_file(get_path_training_data() / "training_data_aggregated") as path:
        if (
            path.joinpath("training_data_" + figure_of_merit + ".npy").is_file()
            and path.joinpath("names_list_" + figure_of_merit + ".npy").is_file()
            and path.joinpath("scores_list_" + figure_of_merit + ".npy").is_file()
        ):
            training_data = np.load(path / ("training_data_" + figure_of_merit + ".npy"), allow_pickle=True)
            names_list = list(np.load(path / ("names_list_" + figure_of_merit + ".npy"), allow_pickle=True))
            scores_list = list(np.load(path / ("scores_list_" + figure_of_merit + ".npy"), allow_pickle=True))
        else:
            error_msg = "Training data not found. Please run the training script first."
            raise FileNotFoundError(error_msg)

        return training_data, names_list, scores_list


@dataclass
class TrainingData:
    X_train: NDArray[np.float_]
    X_test: NDArray[np.float_]
    y_train: NDArray[np.float_]
    y_test: NDArray[np.float_]
    indices_train: list[int]
    indices_test: list[int]
    names_list: list[str]
    scores_list: list[list[float]]



from __future__ import annotations

import json
import sys

if sys.version_info < (3, 10, 0):
    import importlib_resources as resources
else:
    from importlib import resources  # type: ignore[no-redef]
from typing import TypedDict, cast

import numpy as np
from qiskit.providers.fake_provider import FakeMontreal, FakeWashington


class Calibration:
    """The Calibration class is used to store calibration data for different devices."""

    def __init__(self) -> None:
        try:
            self.ibm_montreal_calibration = FakeMontreal().properties()
            self.ibm_washington_calibration = FakeWashington().properties()
            self.oqc_lucy_calibration = parse_oqc_calibration_config()
            self.rigetti_m2_calibration = parse_rigetti_calibration_config()
            self.ionq_harmony_calibration = parse_simple_calibration_config("ionq_harmony")
            self.ionq_aria1_calibration = parse_simple_calibration_config("ionq_aria1")
            self.quantinuum_h2_calibration = parse_simple_calibration_config("quantinuum_h2")

        except Exception as e:
            raise RuntimeError("Error in Calibration initialization: " + str(e)) from e


def get_mean_IBM_washington_cx_error() -> float:
    """Returns the mean cx error for the IBM Washington device."""
    cmap: list[list[int]] = FakeWashington().configuration().coupling_map
    backend = FakeWashington().properties()
    nonfaulty_connections = [x for x in cmap if backend.gate_error("cx", x) < 1]

    res: list[float] = []
    for elem in nonfaulty_connections:
        res.append(backend.gate_error("cx", elem))

    return cast(float, np.mean(res))


def get_mean_IBM_montreal_cx_error() -> float:
    """Returns the mean cx error for the IBM Washington device."""
    cmap: list[list[int]] = FakeMontreal().configuration().coupling_map
    backend = FakeMontreal().properties()
    nonfaulty_connections = [x for x in cmap if backend.gate_error("cx", x) < 1]

    res: list[float] = []
    for elem in nonfaulty_connections:
        res.append(backend.gate_error("cx", elem))

    return cast(float, np.mean(res))


class DeviceCalibration(TypedDict):
    """The DeviceCalibration class is used to store calibration data for different devices."""

    backend: str
    avg_1Q: float
    avg_2Q: float


def parse_simple_calibration_config(device: str) -> DeviceCalibration:
    """Parses the calibration data for the given device.

    Args:
        device (str): The name of the device.

    Returns:
        DeviceCalibration: The calibration data for the given device.
    """

    calibration_filename = device + "_calibration.json"
    ref = resources.files("mqt.predictor") / "calibration_files" / calibration_filename
    with ref.open() as f:
        calibration = json.load(f)
    return {
        "backend": device,
        "avg_1Q": calibration["fidelity"]["1Q"].get("mean"),
        "avg_2Q": calibration["fidelity"]["2Q"].get("mean"),
    }


class OQCCalibration(DeviceCalibration):
    """The OQCCalibration class is used to store calibration data for the OQC device."""

    fid_1Q: dict[str, float]
    fid_1Q_readout: dict[str, float]
    fid_2Q: dict[str, float]


def parse_oqc_calibration_config() -> OQCCalibration:
    """Parses the calibration data for the OQC device.

    Returns:
        OQCCalibration: The calibration data for the OQC device.
    """
    ref = resources.files("mqt.predictor") / "calibration_files" / "oqc_lucy_calibration.json"
    with ref.open() as f:
        oqc_lucy_calibration = json.load(f)
    fid_1Q = {}
    fid_1Q_readout = {}
    for elem in oqc_lucy_calibration["oneQubitProperties"]:
        fid_1Q[str(elem)] = oqc_lucy_calibration["oneQubitProperties"][elem]["oneQubitFidelity"][0].get("fidelity")
        fid_1Q_readout[str(elem)] = oqc_lucy_calibration["oneQubitProperties"][elem]["oneQubitFidelity"][1].get(
            "fidelity"
        )
    fid_2Q = {}
    for elem in oqc_lucy_calibration["twoQubitProperties"]:
        fid_2Q[str(elem)] = oqc_lucy_calibration["twoQubitProperties"][elem]["twoQubitGateFidelity"][0].get("fidelity")

    avg_1Q = np.average(list(fid_1Q.values()))
    avg_2Q = np.average(list(fid_2Q.values()))
    return {
        "backend": "oqc_lucy",
        "avg_1Q": avg_1Q,
        "fid_1Q": fid_1Q,
        "fid_1Q_readout": fid_1Q_readout,
        "avg_2Q": avg_2Q,
        "fid_2Q": fid_2Q,
    }


class RigettiCalibration(DeviceCalibration):
    """The RigettiCalibration class is used to store calibration data for the Rigetti Aspen M2 device."""

    fid_1Q: dict[str, float]
    fid_1Q_readout: dict[str, float]
    fid_2Q_CZ: dict[str, float]


def parse_rigetti_calibration_config() -> RigettiCalibration:
    """Parses the calibration data for the Rigetti Aspen M2 device.

    Returns:
        RigettiCalibration: The calibration data for the Rigetti Aspen M2 device.
    """
    ref = resources.files("mqt.predictor") / "calibration_files" / "rigetti_m2_calibration.json"
    with ref.open() as f:
        rigetti_m2_calibration = json.load(f)
    fid_1Q = {}
    fid_1Q_readout = {}
    missing_indices: list[int] = []
    for elem in rigetti_m2_calibration["specs"]["1Q"]:
        fid_1Q[str(elem)] = rigetti_m2_calibration["specs"]["1Q"][elem].get("f1QRB")
        fid_1Q_readout[str(elem)] = rigetti_m2_calibration["specs"]["1Q"][elem].get("fRO")

    fid_2Q_CZ = {}
    non_list = []
    for elem in rigetti_m2_calibration["specs"]["2Q"]:
        if rigetti_m2_calibration["specs"]["2Q"][elem].get("fCZ") is None:
            non_list.append(elem)
        else:
            fid_2Q_CZ[str(elem)] = rigetti_m2_calibration["specs"]["2Q"][elem].get("fCZ")

    cz_fid_avg = np.average(list(fid_2Q_CZ.values()))

    avg_1Q = np.average(list(fid_1Q.values()))
    for elem in missing_indices:
        fid_2Q_CZ[elem] = cz_fid_avg

    return {
        "backend": "rigetti_aspen_m2",
        "avg_1Q": avg_1Q,
        "fid_1Q": fid_1Q,
        "fid_1Q_readout": fid_1Q_readout,
        "avg_2Q": cz_fid_avg,
        "fid_2Q_CZ": fid_2Q_CZ,
    }



from pathlib import Path
from typing import Final

import numpy as np
from mqt.bench import benchmark_generator
from mqt.predictor import ml, reward


def test_predict() -> None:
    path = ml.helper.get_path_trained_model(figure_of_merit="expected_fidelity")
    assert path.is_file()
    filename = "test_qasm.qasm"
    figure_of_merit: Final = "expected_fidelity"
    qc = benchmark_generator.get_benchmark("dj", 1, 8)
    qc.qasm(filename=filename)
    predictor = ml.Predictor()
    predictions = predictor.predict_probs(filename, figure_of_merit=figure_of_merit)
    assert predictor.clf is not None
    classes = predictor.clf.classes_  # type: ignore[unreachable]
    predicted_device_indices = classes[np.argsort(predictions)[::-1]]
    assert all(0 <= i < len(ml.helper.get_index_to_device_LUT()) for i in predicted_device_indices)
    predictions = predictor.predict_probs(qc.qasm(), figure_of_merit=figure_of_merit)
    predicted_device_indices = classes[np.argsort(predictions)[::-1]]
    assert all(0 <= i < len(ml.helper.get_index_to_device_LUT()) for i in predicted_device_indices)
    Path(filename).unlink()


def test_train_random_forest_classifier() -> None:
    predictor = ml.Predictor()
    assert predictor.clf is None
    predictor.train_random_forest_classifier(visualize_results=False)

    assert predictor.clf is not None


def test_compile_all_circuits_for_dev_and_fom() -> None:
    predictor = ml.Predictor()
    source_path = Path()
    target_path = Path("test_compiled_circuits")
    if not target_path.exists():
        target_path.mkdir()
    figure_of_merit: reward.figure_of_merit = "expected_fidelity"

    qc = benchmark_generator.get_benchmark("dj", 1, 3)
    qasm_path = Path("test.qasm")
    qc.qasm(filename=str(qasm_path))
    predictor.compile_all_circuits_devicewise(
        device_name="ionq_harmony",
        timeout=100,
        figure_of_merit=figure_of_merit,
        source_path=source_path,
        target_path=target_path,
    )
    assert any(file.suffix == ".qasm" for file in target_path.iterdir())

    training_sample, circuit_name, scores = predictor.generate_training_sample(
        file=qasm_path,
        figure_of_merit=figure_of_merit,
        path_uncompiled_circuit=source_path,
        path_compiled_circuits=target_path,
    )
    assert training_sample
    assert circuit_name is not None
    assert any(score != -1 for score in scores)

    (
        training_data,
        name_list,
        scores_list,
    ) = predictor.generate_trainingdata_from_qasm_files(figure_of_merit, source_path, target_path)
    assert len(training_data) > 0
    assert len(name_list) > 0
    assert len(scores_list) > 0

    if target_path.exists():
        for file in target_path.iterdir():
            file.unlink()
        target_path.rmdir()

    if qasm_path.exists():
        qasm_path.unlink()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Literal, cast

import numpy as np
from mqt.bench.utils import calc_qubit_index, calc_supermarq_features
from mqt.predictor import Calibration
from mqt.predictor.utils import (
    get_rigetti_qubit_dict,
)

if TYPE_CHECKING:
    from qiskit import QuantumCircuit

logger = logging.getLogger("mqt-predictor")

figure_of_merit = Literal["expected_fidelity", "critical_depth"]


def crit_depth(qc: QuantumCircuit, precision: int = 10) -> float:
    """Calculates the critical depth of a given quantum circuit."""
    supermarq_features = calc_supermarq_features(qc)
    return cast(float, np.round(1 - supermarq_features.critical_depth, precision))


def expected_fidelity(qc: QuantumCircuit, device_name: str, precision: int = 10) -> float:
    """Calculates the expected fidelity of a given quantum circuit on a given device.

    Args:
        qc (QuantumCircuit): The quantum circuit to be compiled.
        device_name (str): The device to be used for compilation.
        precision (int, optional): The precision of the returned value. Defaults to 10.

    Returns:
        float: The expected fidelity of the given quantum circuit on the given device.
    """

    if "ibm" in device_name:
        res = calc_expected_fidelity_ibm(qc, device_name)

    elif "oqc_lucy" in device_name:
        res = calc_expected_fidelity_oqc_lucy(qc)

    elif "ionq" in device_name:
        res = calc_expected_fidelity_ionq(qc, device_name)

    elif "quantinuum_h2" in device_name:
        res = calc_expected_fidelity_quantinuum_h2(qc)

    elif "rigetti_aspen_m2" in device_name:
        res = calc_expected_fidelity_rigetti_aspen_m2(qc)

    else:
        error_msg = "Device not supported"
        raise ValueError(error_msg)

    return cast(float, np.round(res, precision))


def calc_expected_fidelity_rigetti_aspen_m2(qc: QuantumCircuit) -> float:
    res = 1.0
    calibration = Calibration.Calibration()

    mapping = get_rigetti_qubit_dict()
    for instruction, qargs, _cargs in qc.data:
        gate_type = instruction.name

        assert gate_type in ["rx", "rz", "cz", "measure", "barrier"]
        if gate_type != "barrier":
            assert len(qargs) in [1, 2]
            first_qubit = calc_qubit_index(qargs, qc.qregs, 0)
            if len(qargs) == 1:
                if gate_type == "measure":
                    specific_fidelity = calibration.rigetti_m2_calibration["fid_1Q_readout"][mapping[str(first_qubit)]]
                else:
                    specific_fidelity = calibration.rigetti_m2_calibration["fid_1Q"][mapping[str(first_qubit)]]
            else:
                second_qubit = calc_qubit_index(qargs, qc.qregs, 1)
                tmp = (
                    str(
                        min(
                            int(mapping[str(first_qubit)]),
                            int(mapping[str(second_qubit)]),
                        )
                    )
                    + "-"
                    + str(
                        max(
                            int(mapping[str(first_qubit)]),
                            int(mapping[str(second_qubit)]),
                        )
                    )
                )
                if (
                    calibration.rigetti_m2_calibration["fid_2Q_CZ"].get(tmp) is None
                    or calibration.rigetti_m2_calibration["fid_2Q_CZ"][tmp] is None
                ):
                    specific_fidelity = calibration.rigetti_m2_calibration["avg_2Q"]
                else:
                    specific_fidelity = calibration.rigetti_m2_calibration["fid_2Q_CZ"][tmp]

            res *= specific_fidelity
    return res


def calc_expected_fidelity_quantinuum_h2(qc: QuantumCircuit) -> float:
    res = 1.0
    calibration = Calibration.Calibration().quantinuum_h2_calibration
    for instruction, qargs, _cargs in qc.data:
        gate_type = instruction.name
        assert gate_type in ["rzz", "rz", "ry", "rx", "measure", "barrier"]
        if gate_type != "barrier":
            assert len(qargs) in [1, 2]
            if len(qargs) == 1:
                specific_fidelity = calibration["avg_1Q"]
            elif len(qargs) == 2:
                specific_fidelity = calibration["avg_2Q"]
            res *= specific_fidelity
    return res


def calc_expected_fidelity_ionq(qc: QuantumCircuit, device_name: str) -> float:
    if device_name == "ionq_harmony":
        calibration_data = Calibration.Calibration().ionq_harmony_calibration
    elif device_name == "ionq_aria1":
        calibration_data = Calibration.Calibration().ionq_aria1_calibration
    else:
        msg = "Device not supported"
        raise ValueError(msg)

    res = 1.0
    for instruction, qargs, _cargs in qc.data:
        gate_type = instruction.name

        assert gate_type in ["rxx", "rz", "ry", "rx", "measure", "barrier"]
        if gate_type != "barrier":
            assert len(qargs) in [1, 2]
            if len(qargs) == 1:
                specific_fidelity = calibration_data["avg_1Q"]
            elif len(qargs) == 2:
                specific_fidelity = calibration_data["avg_2Q"]
            res *= specific_fidelity
    return res


def calc_expected_fidelity_oqc_lucy(qc: QuantumCircuit) -> float:
    res = 1.0
    calibration = Calibration.Calibration().oqc_lucy_calibration
    for instruction, qargs, _cargs in qc.data:
        gate_type = instruction.name

        assert gate_type in ["rz", "sx", "x", "ecr", "measure", "barrier"]
        if gate_type != "barrier":
            assert len(qargs) in [1, 2]
            first_qubit = calc_qubit_index(qargs, qc.qregs, 0)
            if len(qargs) == 1 and gate_type != "measure":
                specific_fidelity = calibration["fid_1Q"][str(first_qubit)]
            elif len(qargs) == 1 and gate_type == "measure":
                specific_fidelity = calibration["fid_1Q_readout"][str(first_qubit)]
            elif len(qargs) == 2:
                second_qubit = calc_qubit_index(qargs, qc.qregs, 1)
                tmp = str(first_qubit) + "-" + str(second_qubit)
                if calibration["fid_2Q"].get(tmp) is None:
                    specific_fidelity = calibration["avg_2Q"]
                else:
                    specific_fidelity = calibration["fid_2Q"][tmp]

            res *= specific_fidelity
    return res


def calc_expected_fidelity_ibm(qc: QuantumCircuit, device_name: str) -> float:
    if device_name == "ibm_montreal":
        calibration = Calibration.Calibration().ibm_montreal_calibration
    elif device_name == "ibm_washington":
        calibration = Calibration.Calibration().ibm_washington_calibration
    else:
        msg = "Device not supported"
        raise ValueError(msg)

    res = 1.0
    for instruction, qargs, _cargs in qc.data:
        gate_type = instruction.name

        assert gate_type in ["rz", "sx", "x", "cx", "measure", "barrier"]

        if gate_type != "barrier":
            assert len(qargs) in [1, 2]
            first_qubit = calc_qubit_index(qargs, qc.qregs, 0)
            if len(qargs) == 1:
                try:
                    if gate_type == "measure":
                        specific_error: float = calibration.readout_error(first_qubit)
                    else:
                        specific_error = calibration.gate_error(gate_type, [first_qubit])
                except Exception as e:
                    raise RuntimeError(
                        "Error in IBM backend.gate_error(): "
                        + str(e)
                        + ", "
                        + device_name
                        + ", "
                        + first_qubit
                        + ", "
                        + instruction
                        + ", "
                        + qargs
                    ) from None
            else:
                second_qubit = calc_qubit_index(qargs, qc.qregs, 1)
                try:
                    specific_error = calibration.gate_error(gate_type, [first_qubit, second_qubit])
                    if specific_error == 1:
                        logger.warning(
                            "Gate error is 1.0 for: "
                            + device_name
                            + ", "
                            + gate_type
                            + ", "
                            + str(first_qubit)
                            + ", "
                            + str(second_qubit),
                            "mean error is used instead",
                        )
                        if device_name == "ibm_washington":
                            specific_error = calibration.ibm_washington_cx_mean_error
                        elif device_name == "ibm_montreal":
                            specific_error = calibration.ibm_montreal_cx_mean_error
                except Exception as e:
                    raise RuntimeError(
                        "Error in IBM backend.gate_error(): "
                        + str(e)
                        + ", "
                        + device_name
                        + ", "
                        + first_qubit
                        + ", "
                        + second_qubit
                        + ", "
                        + instruction
                        + ", "
                        + qargs
                    ) from None

            res *= 1 - specific_error
    return res



from mqt.predictor.Calibration import Calibration


def test_calibration() -> None:
    c = Calibration()
    assert c.ibm_montreal_calibration
    assert c.ibm_washington_calibration
    assert c.oqc_lucy_calibration
    assert c.rigetti_m2_calibration
    assert c.ionq_harmony_calibration
    assert c.ionq_aria1_calibration
    assert c.quantinuum_h2_calibration



from __future__ import annotations

import logging
import time
from importlib import resources
from pathlib import Path
from typing import Any

import numpy as np
from joblib import Parallel, delayed
from mqt.bench.tket_helper import get_rebase
from mqt.predictor import Result, ml, reward, rl
from pytket import OpType
from pytket.architecture import Architecture
from pytket.extensions.qiskit import qiskit_to_tk, tk_to_qiskit
from pytket.passes import (
    FullPeepholeOptimise,
    PlacementPass,
    RoutingPass,
)
from pytket.placement import GraphPlacement
from qiskit import QuantumCircuit, transpile

logger = logging.getLogger("mqt-predictor")


def create_qiskit_result(qc: QuantumCircuit, device: dict[str, Any] | None = None) -> Result:
    """Creates a Result object for a given benchmark and device using qiskit for compilation.

    Args:
        benchmark (str): The path to the benchmark to be compiled.
        device (dict[str, Any] | None, optional): The device to be used for compilation. Defaults to None.

    Returns:
        Result: Returns a Result object containing the compiled quantum circuit.
    """
    assert device is not None
    if qc.num_qubits > device["max_qubits"]:
        return Result("qiskit_", -1, None, device["name"])
    start_time = time.time()
    try:
        transpiled_qc_qiskit = transpile(
            qc,
            basis_gates=device["native_gates"],
            coupling_map=device["cmap"],
            optimization_level=3,
            seed_transpiler=1,
        )
    except Exception as e:
        logger.warning("qiskit Transpile Error occurred for: " + device["name"] + " " + str(e))
        return Result("qiskit_" + device["name"], -1, None, device["name"])
    duration = time.time() - start_time
    return Result("qiskit_" + device["name"], duration, transpiled_qc_qiskit, device["name"])


def create_tket_result(
    qc: QuantumCircuit,
    device: dict[str, Any] | None = None,
) -> Result:
    """Creates a Result object for a given benchmark and device using tket for compilation.

    Args:
        benchmark (str): The path to the benchmark to be compiled.
        device (dict[str, Any] | None, optional): The device to be used for compilation. Defaults to None.

    Returns:
        Result: Returns a Result object containing the compiled quantum circuit.
    """
    assert device is not None
    if qc.num_qubits > device["max_qubits"]:
        return Result("tket_" + device["name"], -1, None, device["name"])
    tket_qc = qiskit_to_tk(qc)
    arch = Architecture(device["cmap"])

    native_rebase = get_rebase(device["name"].split("_")[0])
    assert native_rebase is not None

    start_time = time.time()
    try:
        native_rebase.apply(tket_qc)
        FullPeepholeOptimise(target_2qb_gate=OpType.TK2).apply(tket_qc)
        PlacementPass(GraphPlacement(arch)).apply(tket_qc)
        RoutingPass(arch).apply(tket_qc)
        native_rebase.apply(tket_qc)
        duration = time.time() - start_time
        transpiled_qc_tket = tk_to_qiskit(tket_qc)
    except Exception as e:
        logger.warning("tket Transpile Error occurred for: " + device["name"] + " " + str(e))
        return Result("tket_" + device["name"], -1, None, device["name"])

    return Result("tket_" + device["name"], duration, transpiled_qc_tket, device["name"])


def create_mqtpredictor_result(qc: QuantumCircuit, figure_of_merit: reward.figure_of_merit, filename: str) -> Result:
    dev_name = ml.helper.predict_device_for_figure_of_merit(qc, figure_of_merit)
    """ Creates a Result object for a given benchmark and figure of merit using mqt-predictor for compilation.

    Args:
        benchmark (str): The path to the benchmark to be compiled.
        figure_of_merit (reward.reward_functions): The figure of merit to be used for compilation.

    Returns:
        Result: Returns a Result object containing the compiled quantum circuit.
    """
    assert isinstance(dev_name, str)
    dev_index = rl.helper.get_device_index_of_device(dev_name)
    target_filename = filename.split("/")[-1].split(".qasm")[0] + "_" + figure_of_merit + "_" + str(dev_index)
    combined_path_filename = ml.helper.get_path_training_circuits_compiled() / (target_filename + ".qasm")
    if Path(combined_path_filename).exists():
        qc = QuantumCircuit.from_qasm_file(combined_path_filename)
        if qc:
            return Result(
                "mqt-predictor_" + figure_of_merit,
                -1,
                qc,
                dev_name,
            )
    else:
        try:
            qc_compiled = ml.qcompile(qc, figure_of_merit=figure_of_merit)
            if qc_compiled:
                assert isinstance(qc_compiled, tuple)
                return Result(
                    "mqt-predictor_" + figure_of_merit,
                    -1,
                    qc_compiled[0],
                    dev_name,
                )

        except Exception as e:
            logger.warning("mqt-predictor Transpile Error occurred for: " + filename + " " + dev_name + " " + str(e))
    return Result("mqt-predictor_" + figure_of_merit, -1, None, dev_name)


def evaluate_all_sample_circuits() -> None:
    """Evaluates all sample circuits and saves the results to a csv file."""
    res_csv = []

    results = Parallel(n_jobs=-1, verbose=3, backend="threading")(
        delayed(evaluate_sample_circuit)(str(file))
        for file in list(ml.helper.get_path_training_circuits().glob("*.qasm"))
    )
    res_csv.append(list(results[0].keys()))
    for res in results:
        res_csv.append(list(res.values()))
    np.savetxt(
        ml.helper.get_path_results(),
        res_csv,
        delimiter=",",
        fmt="%s",
    )


def evaluate_GHZ_circuits() -> None:
    """Evaluates all GHZ circuits and saves the results to a csv file."""
    res_csv = []

    path = Path(str(resources.files("mqt.predictor"))) / "ml" / "training_data" / "ghz"
    results = Parallel(n_jobs=-1, verbose=3, backend="threading")(
        delayed(evaluate_sample_circuit)(str(file)) for file in list(path.glob("*.qasm"))
    )
    res_csv.append(list(results[0].keys()))
    for res in results:
        res_csv.append(list(res.values()))
    np.savetxt(
        ml.helper.get_path_results(ghz_results=True),
        res_csv,
        delimiter=",",
        fmt="%s",
    )


def evaluate_sample_circuit(filename: str) -> dict[str, Any]:
    """Evaluates a given sample circuit and returns the results as a dictionary.

    Args:
        filename (str): The path to the sample circuit to be evaluated.

    Returns:
        dict[str, Any]: Returns a dictionary containing the results of the evaluation.
    """
    logger.info("Evaluate file: " + filename)

    results: dict[str, Any] = {
        "file_path": str(Path(filename).stem),
        "benchmark_name": str(Path(filename).stem).replace("_", " ").split(" ")[0],
        "num_qubits": str(Path(filename).stem).replace("_", " ").split(" ")[-1],
    }
    qc = QuantumCircuit.from_qasm_file(filename)
    results.update(create_mqtpredictor_result(qc, "expected_fidelity", filename=filename).get_dict())
    results.update(create_mqtpredictor_result(qc, "critical_depth", filename=filename).get_dict())

    for _i, dev in enumerate(rl.helper.get_devices()):
        results.update(create_qiskit_result(qc, dev).get_dict())
        results.update(create_tket_result(qc, dev).get_dict())

    return results



import os

import pytest
from mqt.bench import get_benchmark
from mqt.predictor import reward, rl
from qiskit import QuantumCircuit

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"


@pytest.mark.skipif(
    IN_GITHUB_ACTIONS,
    reason="Trained models are not yet available. Skipping may be reverted after release with trained models.",
)
@pytest.mark.parametrize(
    "figure_of_merit",
    ["expected_fidelity", "critical_depth"],
)
def test_qcompile_with_pretrained_models(figure_of_merit: reward.figure_of_merit) -> None:
    qc = get_benchmark("ghz", 1, 3)
    qc_compiled, compilation_information = rl.qcompile(qc, figure_of_merit=figure_of_merit, device_name="ionq_harmony")
    assert isinstance(qc_compiled, QuantumCircuit)
    assert compilation_information is not None


@pytest.mark.parametrize(
    "figure_of_merit",
    ["expected_fidelity", "critical_depth"],
)
def test_qcompile_with_newly_trained_models(figure_of_merit: reward.figure_of_merit) -> None:
    predictor = rl.Predictor(figure_of_merit=figure_of_merit, device_name="ionq_harmony")
    predictor.train_model(
        timesteps=100,
        test=True,
    )

    qc = get_benchmark("ghz", 1, 5)
    res = rl.qcompile(qc, figure_of_merit=figure_of_merit, device_name="ionq_harmony")
    assert type(res) == tuple
    qc_compiled, compilation_information = res

    assert isinstance(qc_compiled, QuantumCircuit)
    assert compilation_information is not None



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pathlib import Path

import numpy as np
from gymnasium import Env
from gymnasium.spaces import Box, Dict, Discrete
from mqt.predictor import reward, rl
from pytket.extensions.qiskit import qiskit_to_tk, tk_to_qiskit
from qiskit import QuantumCircuit
from qiskit.transpiler import CouplingMap, PassManager
from qiskit.transpiler.passes import CheckMap, GatesInBasis
from qiskit.transpiler.runningpassmanager import TranspileLayout

logger = logging.getLogger("mqt-predictor")


class PredictorEnv(Env):  # type: ignore[misc]
    """Predictor environment for reinforcement learning."""

    def __init__(
        self, reward_function: reward.figure_of_merit = "expected_fidelity", device_name: str = "ibm_washington"
    ):
        logger.info("Init env: " + reward_function)

        self.action_set = {}
        self.actions_synthesis_indices = []
        self.actions_layout_indices = []
        self.actions_routing_indices = []
        self.actions_mapping_indices = []
        self.actions_opt_indices = []
        self.used_actions: list[str] = []
        self.device = rl.helper.get_device(device_name)

        index = 0

        for elem in rl.helper.get_actions_synthesis():
            self.action_set[index] = elem
            self.actions_synthesis_indices.append(index)
            index += 1
        for elem in rl.helper.get_actions_layout():
            self.action_set[index] = elem
            self.actions_layout_indices.append(index)
            index += 1
        for elem in rl.helper.get_actions_routing():
            self.action_set[index] = elem
            self.actions_routing_indices.append(index)
            index += 1
        for elem in rl.helper.get_actions_opt():
            self.action_set[index] = elem
            self.actions_opt_indices.append(index)
            index += 1
        for elem in rl.helper.get_actions_mapping():
            self.action_set[index] = elem
            self.actions_mapping_indices.append(index)
            index += 1

        self.action_set[index] = rl.helper.get_action_terminate()
        self.action_terminate_index = index

        self.reward_function = reward_function
        self.action_space = Discrete(len(self.action_set.keys()))
        self.num_steps = 0
        self.layout = None

        spaces = {
            "num_qubits": Discrete(128),
            "depth": Discrete(1000000),
            "program_communication": Box(low=0, high=1, shape=(1,), dtype=np.float32),
            "critical_depth": Box(low=0, high=1, shape=(1,), dtype=np.float32),
            "entanglement_ratio": Box(low=0, high=1, shape=(1,), dtype=np.float32),
            "parallelism": Box(low=0, high=1, shape=(1,), dtype=np.float32),
            "liveness": Box(low=0, high=1, shape=(1,), dtype=np.float32),
        }
        self.observation_space = Dict(spaces)
        self.filename = ""

    def step(self, action: int) -> tuple[dict[str, Any], float, bool, bool, dict[Any, Any]]:
        """Executes the given action and returns the new state, the reward, whether the episode is done, whether the episode is truncated and additional information."""
        self.used_actions.append(str(self.action_set[action].get("name")))
        altered_qc = self.apply_action(action)
        if not altered_qc:
            return (
                rl.helper.create_feature_dict(self.state),
                0,
                True,
                False,
                {},
            )

        self.state: QuantumCircuit = altered_qc
        self.num_steps += 1

        self.valid_actions = self.determine_valid_actions_for_state()
        if len(self.valid_actions) == 0:
            msg = "No valid actions left."
            raise RuntimeError(msg)

        if action == self.action_terminate_index:
            reward_val = self.calculate_reward()
            done = True
        else:
            reward_val = 0
            done = False

        # in case the Qiskit.QuantumCircuit has unitary or u gates in it, decompose them (because otherwise qiskit will throw an error when applying the BasisTranslator
        if self.state.count_ops().get("unitary"):
            self.state = self.state.decompose(gates_to_decompose="unitary")

        obs = rl.helper.create_feature_dict(self.state)
        return obs, reward_val, done, False, {}

    def calculate_reward(self) -> Any:
        """Calculates and returns the reward for the current state."""
        if self.reward_function == "expected_fidelity":
            return reward.expected_fidelity(self.state, self.device["name"])
        if self.reward_function == "critical_depth":
            return reward.crit_depth(self.state)
        error_msg = f"Reward function {self.reward_function} not supported."  # type: ignore[unreachable]
        raise ValueError(error_msg)

    def render(self) -> None:
        """Renders the current state."""
        print(self.state.draw())

    def reset(
        self,
        qc: Path | str | QuantumCircuit | None = None,
        seed: int | None = None,
        options: dict[str, Any] | None = None,  # noqa: ARG002
    ) -> tuple[QuantumCircuit, dict[str, Any]]:
        """Resets the environment to the given state or a random state.

        Args:
            qc (Path | str | QuantumCircuit | None, optional): The quantum circuit to be compiled or the path to a qasm file containing the quantum circuit. Defaults to None.
            seed (int | None, optional): The seed to be used for the random number generator. Defaults to None.
            options (dict[str, Any] | None, optional): Additional options. Defaults to None.

        Returns:
            tuple[QuantumCircuit, dict[str, Any]]: The initial state and additional information.
        """
        super().reset(seed=seed)
        if isinstance(qc, QuantumCircuit):
            self.state = qc
        elif qc:
            self.state = QuantumCircuit.from_qasm_file(str(qc))
        else:
            self.state, self.filename = rl.helper.get_state_sample(self.device["max_qubits"])

        self.action_space = Discrete(len(self.action_set.keys()))
        self.num_steps = 0
        self.used_actions = []

        self.layout = None

        self.valid_actions = self.actions_opt_indices + self.actions_synthesis_indices

        self.error_occured = False
        return rl.helper.create_feature_dict(self.state), {}

    def action_masks(self) -> list[bool]:
        """Returns a list of valid actions for the current state."""
        return [action in self.valid_actions for action in self.action_set]

    def apply_action(self, action_index: int) -> QuantumCircuit | None:
        """Applies the given action to the current state and returns the altered state."""
        if action_index in self.action_set:
            action = self.action_set[action_index]
            if action["name"] == "terminate":
                return self.state
            if (
                action_index
                in self.actions_layout_indices + self.actions_routing_indices + self.actions_mapping_indices
            ):
                transpile_pass = action["transpile_pass"](self.device["cmap"])
            elif action_index in self.actions_synthesis_indices:
                transpile_pass = action["transpile_pass"](self.device["native_gates"])
            else:
                transpile_pass = action["transpile_pass"]
            if action["origin"] == "qiskit":
                try:
                    if action["name"] == "QiskitO3":
                        pm = PassManager()
                        pm.append(
                            action["transpile_pass"](self.device["native_gates"], CouplingMap(self.device["cmap"])),
                            do_while=action["do_while"],
                        )
                    else:
                        pm = PassManager(transpile_pass)
                    altered_qc = pm.run(self.state)
                except Exception as e:
                    logger.error(
                        "Error in executing Qiskit transpile pass for {action} at step {i} for {filename}: {e}".format(
                            action=action["name"], i=self.num_steps, filename=self.filename, e=e
                        )
                    )

                    self.error_occured = True
                    return None
                if action_index in self.actions_layout_indices + self.actions_mapping_indices:
                    assert pm.property_set["layout"]
                    self.layout = TranspileLayout(
                        initial_layout=pm.property_set["layout"],
                        input_qubit_mapping=pm.property_set["original_qubit_indices"],
                        final_layout=pm.property_set["final_layout"],
                    )

            elif action["origin"] == "tket":
                try:
                    tket_qc = qiskit_to_tk(self.state)
                    for elem in transpile_pass:
                        elem.apply(tket_qc)
                    altered_qc = tk_to_qiskit(tket_qc)
                except Exception as e:
                    logger.error(
                        "Error in executing TKET transpile  pass for {action} at step {i} for {filename}: {e}".format(
                            action=action["name"], i=self.num_steps, filename=self.filename, e=e
                        )
                    )
                    self.error_occured = True
                    return None

            else:
                error_msg = f"Origin {action['origin']} not supported."
                raise ValueError(error_msg)

        else:
            error_msg = f"Action {action_index} not supported."
            raise ValueError(error_msg)

        return altered_qc

    def determine_valid_actions_for_state(self) -> list[int]:
        """Determines and returns the valid actions for the current state."""
        check_nat_gates = GatesInBasis(basis_gates=self.device["native_gates"])
        check_nat_gates(self.state)
        only_nat_gates = check_nat_gates.property_set["all_gates_in_basis"]

        if not only_nat_gates:
            return self.actions_synthesis_indices + self.actions_opt_indices

        check_mapping = CheckMap(coupling_map=CouplingMap(self.device["cmap"]))
        check_mapping(self.state)
        mapped = check_mapping.property_set["is_swap_mapped"]

        if mapped and self.layout is not None:
            return [self.action_terminate_index, *self.actions_opt_indices]  # type: ignore[unreachable]

        if self.state._layout is not None:
            return self.actions_routing_indices

        # No layout applied yet
        return self.actions_mapping_indices + self.actions_layout_indices + self.actions_opt_indices



from __future__ import annotations

from typing import TYPE_CHECKING

from mqt.predictor import reward

if TYPE_CHECKING:
    from qiskit import QuantumCircuit


class Result:
    """
    The Result class is used to store the results of a compilation.

    Attributes:
        compilation_setup (str): The setup used for compilation. Either 'mqt-predictor_<figure_of_merit>', 'qiskit' or 'tket'. For the two latter, also the device name is appended.
        compilation_time (float): The time it took to compile the benchmark.
        compiled_qc (QuantumCircuit | None): The compiled quantum circuit. If compilation failed, None is returned.
        device (str): The device used for compilation.

    """

    def __init__(
        self,
        compilation_setup: str,
        compilation_time: float,
        compiled_qc: QuantumCircuit | None,
        device: str,
    ) -> None:
        if compiled_qc is not None:
            rew_fid = reward.expected_fidelity(compiled_qc, device)
            rew_crit_depth = reward.crit_depth(compiled_qc)
        else:
            rew_fid = -1.0
            rew_crit_depth = -1.0

        self.compiler = compilation_setup
        self.compilation_time = compilation_time
        self.fidelity = rew_fid
        self.critical_depth = rew_crit_depth

    def get_dict(self) -> dict[str, float]:
        """Returns the results as a dictionary."""

        return {
            self.compiler + "_" + "time": self.compilation_time,
            self.compiler + "_" + "expected_fidelity": self.fidelity,
            self.compiler + "_" + "critical_depth": self.critical_depth,
        }



import logging
from mqt.predictor.Result import Result
from mqt.predictor.ml import qcompile

__all__ = [
    "Result",
    "qcompile",
]

logger = logging.getLogger("mqt-predictor")

console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
logger_formatter = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
console_handler.setFormatter(logger_formatter)
logger.addHandler(console_handler)
logger.setLevel(logging.DEBUG)



from mqt.bench import benchmark_generator
from mqt.predictor import ml


def test_get_index_to_device_LUT() -> None:
    expected = {
        0: "ibm_washington",
        1: "ibm_montreal",
        2: "oqc_lucy",
        3: "rigetti_aspen_m2",
        4: "ionq_harmony",
        5: "ionq_aria1",
        6: "quantinuum_h2",
    }
    assert ml.helper.get_index_to_device_LUT() == expected


def test_load_training_data() -> None:
    assert ml.helper.load_training_data() is not None


def test_create_feature_dict() -> None:
    qc = benchmark_generator.get_benchmark("dj", 1, 3)
    feature_vector = ml.helper.create_feature_dict(qc)
    assert feature_vector is not None


def test_get_openqasm_gates() -> None:
    assert ml.helper.get_openqasm_gates() is not None


def test_get_path_training_circuits() -> None:
    path = ml.helper.get_path_training_circuits()
    assert path.exists()


def test_get_path_training_circuits_compiled() -> None:
    path = ml.helper.get_path_training_circuits_compiled()
    assert path.exists()


def test_get_path_training_data() -> None:
    path = ml.helper.get_path_training_data()
    assert path.exists()


def test_get_path_trained_model() -> None:
    for figure_of_merit in ["expected_fidelity", "critical_depth"]:
        path = ml.helper.get_path_trained_model(figure_of_merit=figure_of_merit)
        assert path.exists()


def test_predict_device_for_figure_of_merit() -> None:
    qc = benchmark_generator.get_benchmark("ghz", 1, 5)
    assert (
        ml.helper.predict_device_for_figure_of_merit(qc, "expected_fidelity")
        in ml.helper.get_index_to_device_LUT().values()
    )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from mqt.predictor import reward, rl
from sb3_contrib import MaskablePPO
from sb3_contrib.common.maskable.policies import MaskableMultiInputActorCriticPolicy
from sb3_contrib.common.maskable.utils import get_action_masks

if TYPE_CHECKING:
    from qiskit import QuantumCircuit

logger = logging.getLogger("mqt-predictor")
PATH_LENGTH = 260


class Predictor:
    """The Predictor class is used to train a reinforcement learning model for a given figure of merit and device such that it acts as a compiler."""

    def __init__(self, figure_of_merit: reward.figure_of_merit, device_name: str, logger_level: int = logging.INFO):
        logger.setLevel(logger_level)

        self.model = None
        self.env = rl.PredictorEnv(figure_of_merit, device_name)
        self.device_name = device_name
        self.figure_of_merit = figure_of_merit

    def load_model(self) -> None:
        self.model = rl.helper.load_model("model_" + self.figure_of_merit + "_" + self.device_name)

    def compile_as_predicted(
        self,
        qc: QuantumCircuit,
    ) -> tuple[QuantumCircuit, list[str]]:
        """Compiles a given quantum circuit such that the given figure of merit is maximized by using the respectively trained optimized compiler.

        Args:
            qc (QuantumCircuit: The quantum circuit to be compiled or the path to a qasm file containing the quantum circuit.

        Returns:
            tuple[QuantumCircuit, list[str]] | bool: Returns a tuple containing the compiled quantum circuit and the compilation information. If compilation fails, False is returned.
        """
        if not self.model:
            try:
                self.load_model()
            except Exception as e:
                msg = "Model cannot be loaded. Try to train a local model using 'Predictor.train_model(<...>)'"
                raise Exception(msg) from e

        assert self.model
        obs, _ = self.env.reset(qc)  # type: ignore[unreachable]

        used_compilation_passes = []
        terminated = False
        truncated = False
        while not (terminated or truncated):
            action_masks = get_action_masks(self.env)
            action, _ = self.model.predict(obs, action_masks=action_masks)
            action = int(action)
            action_item = self.env.action_set[action]
            used_compilation_passes.append(action_item["name"])
            obs, reward_val, terminated, truncated, info = self.env.step(action)
            self.env.state._layout = self.env.layout

        if not self.env.error_occured:
            return self.env.state, used_compilation_passes

        msg = "Error occurred during compilation."
        raise Exception(msg)

    def train_model(
        self,
        timesteps: int = 1000,
        model_name: str = "model",
        verbose: int = 2,
        test: bool = False,
    ) -> None:
        """Trains all models for the given reward functions and device.

        Args:
            timesteps (int, optional): The number of timesteps to train the model. Defaults to 1000.
            model_name (str, optional): The name of the model. Defaults to "model".
            verbose (int, optional): The verbosity level. Defaults to 2.
            test (bool, optional): Whether to train the model for testing purposes. Defaults to False.
        """

        if test:
            n_steps = 100
            progress_bar = False
        else:
            n_steps = 2048
            progress_bar = True

        logger.debug("Start training for: " + self.figure_of_merit + " on " + self.device_name)
        env = rl.PredictorEnv(reward_function=self.figure_of_merit, device_name=self.device_name)

        model = MaskablePPO(
            MaskableMultiInputActorCriticPolicy,
            env,
            verbose=verbose,
            tensorboard_log="./" + model_name + "_" + self.figure_of_merit + "_" + self.device_name,
            gamma=0.98,
            n_steps=n_steps,
        )
        model.learn(total_timesteps=timesteps, progress_bar=progress_bar)
        model.save(
            rl.helper.get_path_trained_model() / (model_name + "_" + self.figure_of_merit + "_" + self.device_name)
        )



from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import matplotlib.pyplot as plt
import numpy as np
from joblib import Parallel, delayed, load
from mqt.predictor import ml, reward, rl, utils
from qiskit import QuantumCircuit
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import GridSearchCV, train_test_split

if TYPE_CHECKING:
    from numpy._typing import NDArray

plt.rcParams["font.family"] = "Times New Roman"

logger = logging.getLogger("mqt-predictor")


class Predictor:
    def __init__(self, logger_level: int = logging.INFO) -> None:
        logger.setLevel(logger_level)

        self.clf = None

    def set_classifier(self, clf: RandomForestClassifier) -> None:
        """Sets the classifier to the given classifier"""
        self.clf = clf

    def compile_all_circuits_circuitwise(
        self,
        figure_of_merit: reward.figure_of_merit,
        timeout: int,
        source_path: Path | None = None,
        target_path: Path | None = None,
        logger_level: int = logging.INFO,
    ) -> None:
        """Compiles all circuits in the given directory with the given timeout and saves them in the given directory.

        Args:
            timeout (int): The timeout in seconds for the compilation of a single circuit
            source_path (Path, optional): The path to the directory containing the circuits to be compiled. Defaults to None.
            target_path (Path, optional): The path to the directory where the compiled circuits should be saved. Defaults to None.
            logger_level (int, optional): The level of the logger. Defaults to logging.INFO.

        """
        logger.setLevel(logger_level)

        if source_path is None:
            source_path = ml.helper.get_path_training_circuits()

        if target_path is None:
            target_path = ml.helper.get_path_training_circuits_compiled()

        Parallel(n_jobs=-1, verbose=100)(
            delayed(self.generate_compiled_circuits_for_single_training_circuit)(
                filename, timeout, source_path, target_path, figure_of_merit
            )
            for filename in source_path.iterdir()
        )

    def generate_compiled_circuits_for_single_training_circuit(
        self,
        filename: Path,
        timeout: int,
        source_path: Path,
        target_path: Path,
        figure_of_merit: reward.figure_of_merit,
    ) -> None:
        """Compiles a single circuit with the given timeout and saves it in the given directory.

        Args:
            filename (Path): The path to the circuit to be compiled
            timeout (int): The timeout in seconds for the compilation of the circuit
            source_path (Path): The path to the directory containing the circuit to be compiled
            target_path (Path): The path to the directory where the compiled circuit should be saved
            figure_of_merit (reward.figure_of_merit): The figure of merit to be used for compilation.

        """
        try:
            qc = QuantumCircuit.from_qasm_file(Path(source_path) / filename)
            if filename.suffix != ".qasm":
                return

            for i, dev in enumerate(rl.helper.get_devices()):
                target_filename = str(filename).split("/")[-1].split(".qasm")[0] + "_" + figure_of_merit + "_" + str(i)
                if (Path(target_path) / (target_filename + ".qasm")).exists() or qc.num_qubits > dev["max_qubits"]:
                    continue
                try:
                    res = utils.timeout_watcher(rl.qcompile, [qc, figure_of_merit, dev["name"]], timeout)
                    if res:
                        compiled_qc = res[0]
                        compiled_qc.qasm(filename=Path(target_path) / (target_filename + ".qasm"))

                except Exception as e:
                    print(e, filename, "inner")

        except Exception as e:
            print(e, filename, "outer")

    def compile_all_circuits_devicewise(
        self,
        device_name: str,
        timeout: int,
        figure_of_merit: reward.figure_of_merit,
        source_path: Path | None = None,
        target_path: Path | None = None,
        logger_level: int = logging.INFO,
    ) -> None:
        """Compiles all circuits in the given directory with the given timeout and saves them in the given directory.

        Args:
            device_name (str): The name of the device to be used for compilation
            timeout (int): The timeout in seconds for the compilation of a single circuit
            figure_of_merit (reward.reward_functions): The figure of merit to be used for compilation
            source_path (Path, optional): The path to the directory containing the circuits to be compiled. Defaults to None.
            target_path (Path, optional): The path to the directory where the compiled circuits should be saved. Defaults to None.
            logger_level (int, optional): The level of the logger. Defaults to logging.INFO.
        """
        logger.setLevel(logger_level)

        logger.info("Processing: " + device_name + " for " + figure_of_merit)
        rl_pred = rl.Predictor(figure_of_merit=figure_of_merit, device_name=device_name)

        dev_index = rl.helper.get_device_index_of_device(device_name)
        dev_max_qubits = rl.helper.get_devices()[dev_index]["max_qubits"]

        if source_path is None:
            source_path = ml.helper.get_path_training_circuits()

        if target_path is None:
            target_path = ml.helper.get_path_training_circuits_compiled()

        for filename in source_path.iterdir():
            if filename.suffix != ".qasm":
                continue
            qc = QuantumCircuit.from_qasm_file(Path(source_path) / filename)
            if qc.num_qubits > dev_max_qubits:
                continue

            target_filename = (
                str(filename).split("/")[-1].split(".qasm")[0] + "_" + figure_of_merit + "_" + str(dev_index)
            )
            if (Path(target_path) / (target_filename + ".qasm")).exists():
                continue
            try:
                res = utils.timeout_watcher(rl.qcompile, [qc, figure_of_merit, device_name, rl_pred], timeout)
                if res:
                    compiled_qc = res[0]
                    compiled_qc.qasm(filename=Path(target_path) / (target_filename + ".qasm"))

            except Exception as e:
                print(e, filename, device_name)
                raise RuntimeError("Error during compilation: " + str(e)) from e

    def generate_compiled_circuits(
        self,
        source_path: Path | None = None,
        target_path: Path | None = None,
        timeout: int = 600,
    ) -> None:
        """Compiles all circuits in the given directory with the given timeout and saves them in the given directory.

        Args:
            source_path (Path, optional): The path to the directory containing the circuits to be compiled. Defaults to None.
            target_path (Path, optional): The path to the directory where the compiled circuits should be saved. Defaults to None.
            timeout (int, optional): The timeout in seconds for the compilation of a single circuit. Defaults to 600.
        """
        if source_path is None:
            source_path = ml.helper.get_path_training_circuits()

        if target_path is None:
            target_path = ml.helper.get_path_training_circuits_compiled()

        path_zip = source_path / "training_data_device_selection.zip"
        if not any(file.suffix == ".qasm" for file in source_path.iterdir()) and path_zip.exists():
            import zipfile

            with zipfile.ZipFile(str(path_zip), "r") as zip_ref:
                zip_ref.extractall(source_path)

        target_path.mkdir(exist_ok=True)

        Parallel(n_jobs=1, verbose=100)(
            delayed(self.compile_all_circuits_devicewise)(
                device_name, timeout, figure_of_merit, source_path, target_path, logger.level
            )
            for figure_of_merit in ["expected_fidelity", "critical_depth"]
            for device_name in [dev["name"] for dev in rl.helper.get_devices()]
        )

    def generate_trainingdata_from_qasm_files(
        self,
        figure_of_merit: reward.figure_of_merit,
        path_uncompiled_circuits: Path | None = None,
        path_compiled_circuits: Path | None = None,
    ) -> tuple[list[NDArray[np.float_]], list[str], list[NDArray[np.float_]]]:
        """Handles to create training data from all generated training samples

        Args:
            figure_of_merit (reward.reward_functions): The figure of merit to be used for training
            path_uncompiled_circuits (Path, optional): The path to the directory containing the uncompiled circuits. Defaults to None.
            path_compiled_circuits (Path, optional): The path to the directory containing the compiled circuits. Defaults to None.

        Returns:
            tuple[list[Any], list[Any], list[Any]]: The training data, consisting of training_data, name_list, scores_list

        """
        if not path_uncompiled_circuits:
            path_uncompiled_circuits = ml.helper.get_path_training_circuits()

        if not path_compiled_circuits:
            path_compiled_circuits = ml.helper.get_path_training_circuits_compiled()

        # init resulting list (feature vector, name, scores)
        training_data = []
        name_list = []
        scores_list = []

        results = Parallel(n_jobs=-1, verbose=100)(
            delayed(self.generate_training_sample)(
                filename.name,
                path_uncompiled_circuits,
                path_compiled_circuits,
                figure_of_merit,
                logger.level,
            )
            for filename in path_uncompiled_circuits.glob("*.qasm")
        )
        for sample in results:
            training_sample, circuit_name, scores = sample
            if all(score == -1 for score in scores):
                continue
            training_data.append(training_sample)
            name_list.append(circuit_name)
            scores_list.append(scores)

        return (training_data, name_list, scores_list)

    def generate_training_sample(
        self,
        file: Path,
        path_uncompiled_circuit: Path,
        path_compiled_circuits: Path,
        figure_of_merit: reward.figure_of_merit = "expected_fidelity",
        logger_level: int = logging.INFO,
    ) -> tuple[tuple[list[Any], Any], str, list[float]]:
        """Handles to create a training sample from a given file.

        Args:
            file (Path): The name of the file to be used for training
            figure_of_merit (reward.reward_functions, optional): The figure of merit to be used for compilation. Defaults to "expected_fidelity".
            path_uncompiled_circuit (Path): The path to the directory containing the uncompiled circuits. Defaults to None.
            path_compiled_circuits (Path): The path to the directory containing the compiled circuits. Defaults to None.
            logger_level (int, optional): The level of the logger. Defaults to logging.INFO.

        Returns:
            tuple[tuple[list[Any], Any], str, list[float]]: Training_sample, circuit_name, scores
        """
        logger.setLevel(logger_level)

        if ".qasm" not in str(file):
            raise RuntimeError("File is not a qasm file: " + str(file))

        LUT = ml.helper.get_index_to_device_LUT()
        logger.debug("Checking " + str(file))
        scores: list[float] = []
        for _ in range(len(LUT)):
            scores.append(-1.0)
        all_relevant_files = path_compiled_circuits.glob(str(file).split(".")[0] + "*")

        for filename in all_relevant_files:
            filename_str = str(filename)
            if (str(file).split(".")[0] + "_" + figure_of_merit + "_") not in filename_str and filename_str.endswith(
                ".qasm"
            ):
                continue
            comp_path_index = int(filename_str.split("_")[-1].split(".")[0])
            device = LUT[comp_path_index]
            qc = QuantumCircuit.from_qasm_file(filename_str)
            if figure_of_merit == "critical_depth":
                score = reward.crit_depth(qc)
            elif figure_of_merit == "expected_fidelity":
                score = reward.expected_fidelity(qc, device)
            scores[comp_path_index] = score

        num_not_empty_entries = 0
        for i in range(len(LUT)):
            if scores[i] != -1.0:
                num_not_empty_entries += 1

        if num_not_empty_entries == 0:
            logger.warning("no compiled circuits found for:" + str(file))

        feature_vec = ml.helper.create_feature_dict(str(path_uncompiled_circuit / file))
        training_sample = (list(feature_vec.values()), np.argmax(scores))
        circuit_name = str(file).split(".")[0]
        return (training_sample, circuit_name, scores)

    def train_random_forest_classifier(
        self, figure_of_merit: reward.figure_of_merit = "expected_fidelity", visualize_results: bool = False
    ) -> bool:
        """Trains a random forest classifier for the given figure of merit.

        Args:
            figure_of_merit (reward.reward_functions, optional): The figure of merit to be used for training. Defaults to "expected_fidelity".
            visualize_results (bool, optional): Whether to visualize the results. Defaults to False.

        Returns:
            bool: Whether the training was successful.
        """

        training_data = self.get_prepared_training_data(figure_of_merit, save_non_zero_indices=True)

        scores_filtered = [training_data.scores_list[i] for i in training_data.indices_test]
        names_filtered = [training_data.names_list[i] for i in training_data.indices_test]

        tree_param = [
            {
                "n_estimators": [100, 200, 500],
                "max_depth": list(range(8, 30, 6)),
                "min_samples_split": list(range(2, 20, 6)),
                "min_samples_leaf": list(range(2, 20, 6)),
                "bootstrap": [True, False],
            },
        ]

        clf = RandomForestClassifier(random_state=0)
        clf = GridSearchCV(clf, tree_param, cv=5, n_jobs=8).fit(training_data.X_train, training_data.y_train)

        if visualize_results:
            y_pred = np.array(list(clf.predict(training_data.X_test)))
            res, _ = self.calc_performance_measures(scores_filtered, y_pred, training_data.y_test)
            self.plot_eval_histogram(res, filename="RandomForestClassifier")

            logger.info("Best Accuracy: " + str(clf.best_score_))
            top3 = (res.count(1) + res.count(2) + res.count(3)) / len(res)
            logger.info("Top 3: " + str(top3))
            logger.info("Feature Importance: " + str(clf.best_estimator_.feature_importances_))
            self.plot_eval_all_detailed_compact_normed(names_filtered, scores_filtered, y_pred, training_data.y_test)

        self.set_classifier(clf.best_estimator_)
        ml.helper.save_classifier(clf.best_estimator_, figure_of_merit)
        logger.info("Random Forest classifier is trained and saved.")

        return self.clf is not None

    def get_prepared_training_data(
        self, figure_of_merit: reward.figure_of_merit, save_non_zero_indices: bool = False
    ) -> ml.helper.TrainingData:
        """Prepares the training data for the given figure of merit.

        Args:
            figure_of_merit (reward.reward_functions): The figure of merit to be used for training.
            save_non_zero_indices (bool, optional): Whether to save the non zero indices. Defaults to False.

        Returns:
            ml.helper.TrainingData: The prepared training data.
        """
        training_data, names_list, raw_scores_list = ml.helper.load_training_data(figure_of_merit)
        unzipped_training_data_X, unzipped_training_data_Y = zip(*training_data)
        scores_list: list[list[float]] = [[] for _ in range(len(raw_scores_list))]
        X_raw = list(unzipped_training_data_X)
        X_list: list[list[float]] = [[] for _ in range(len(X_raw))]
        y_list = list(unzipped_training_data_Y)
        for i in range(len(X_raw)):
            X_list[i] = list(X_raw[i])
            scores_list[i] = list(raw_scores_list[i])

        X, y, indices = np.array(X_list), np.array(y_list), np.array(range(len(y_list)))

        # Store all non zero feature indices
        non_zero_indices = []
        for i in range(len(X[0])):
            if sum(X[:, i]) > 0:
                non_zero_indices.append(i)
        X = X[:, non_zero_indices]

        if save_non_zero_indices:
            data = np.asarray(non_zero_indices)
            np.save(
                ml.helper.get_path_trained_model(figure_of_merit, return_non_zero_indices=True),
                data,
            )

        (
            X_train,
            X_test,
            y_train,
            y_test,
            indices_train,
            indices_test,
        ) = train_test_split(X, y, indices, test_size=0.3, random_state=5)

        return ml.helper.TrainingData(
            X_train,
            X_test,
            y_train,
            y_test,
            indices_train,
            indices_test,
            names_list,
            scores_list,
        )

    def calc_performance_measures(
        self,
        scores_filtered: list[list[float]],
        y_pred: np.ndarray[Any, np.dtype[np.float64]],
        y_test: np.ndarray[Any, np.dtype[np.float64]],
    ) -> tuple[list[int], list[float]]:
        """Method to generate the performance measures for a trained classifier

        Args:
            scores_filtered (list[list[float]]): The scores filtered for the respectively predicted indices of all training data
            y_pred (np.ndarray[Any, np.dtype[np.float64]]): The predicted labels
            y_test (np.ndarray[Any, np.dtype[np.float64]]): The actual labels

        Returns:
            tuple[list[int], list[float]]: The ranks and the relative scores
        """

        res = []
        relative_scores = []
        for i in range(len(y_pred)):
            assert np.argmax(scores_filtered[i]) == y_test[i]
            predicted_score = scores_filtered[i][y_pred[i]]
            relative_scores.append(predicted_score - np.max(scores_filtered[i]))
            score = list(np.sort(scores_filtered[i])[::-1]).index(predicted_score)
            res.append(score + 1)

        assert len(res) == len(y_pred)

        return res, relative_scores

    def plot_eval_histogram(self, res: list[int], filename: str = "histogram", color: str = "#21918c") -> None:
        """Method to generate the histogram for the evaluation scores

        Args:
            res (list[int]): The ranks of the predictions
            filename (str, optional): The filename of the histogram. Defaults to "histogram".
            color (str, optional): The color of the histogram. Defaults to "#21918c".

        """

        plt.figure(figsize=(10, 5))

        num_of_comp_paths = len(ml.helper.get_index_to_device_LUT())
        plt.bar(
            list(range(0, num_of_comp_paths, 1)),
            height=[res.count(i) / len(res) for i in range(1, num_of_comp_paths + 1, 1)],
            width=0.90,
            color=color,
        )

        plt.xticks(
            list(range(0, num_of_comp_paths, 1)),
            [i if i % 2 == 1 else "" for i in range(1, num_of_comp_paths + 1, 1)],
            fontsize=16,
        )

        plt.yticks(fontsize=16)

        plt.xlabel(
            "Best prediction                                                        Worst prediction",
            fontsize=18,
        )
        plt.ylabel("Relative frequency", fontsize=18)
        result_path = Path("results")
        if not result_path.is_dir():
            result_path.mkdir()
        plt.savefig(result_path / (filename + ".pdf"), bbox_inches="tight")
        plt.show()

    def plot_eval_all_detailed_compact_normed(
        self,
        names_list: list[Any],
        scores_filtered: list[Any],
        y_pred: np.ndarray[Any, np.dtype[np.float64]],
        y_test: np.ndarray[Any, np.dtype[np.float64]],
        color_all: str = "#21918c",
        color_pred: str = "#440154",
    ) -> None:
        """Method to generate the plot for the evaluation scores of all training data

        Args:
            names_list (list[Any]): The names of all training data
            scores_filtered (list[Any]): The scores filtered for the respectively predicted indices of all training data
            y_pred (np.ndarray[Any, np.dtype[np.float64]]): The predicted labels
            y_test (np.ndarray[Any, np.dtype[np.float64]]): The actual labels
            color_all (str, optional): The color of the evaluation scores of all training data. Defaults to "#21918c".
            color_pred (str, optional): The color of the evaluation scores of the predicted training data. Defaults to "#440154".

        """

        # Create list of all qubit numbers and sort them
        names_list_num_qubits = []
        for i in range(len(names_list)):
            assert np.argmax(scores_filtered[i]) == y_test[i]
            names_list_num_qubits.append(int(names_list[i].split("_")[-1].split(".")[0]))

        # Sort all other list (num_qubits, scores and y_pred) accordingly
        (
            qubit_list_sorted,
            scores_filtered_sorted_accordingly,
            y_pred_sorted_accordingly,
        ) = zip(*sorted(zip(names_list_num_qubits, scores_filtered, y_pred)))
        plt.figure(figsize=(17, 8))
        for i in range(len(names_list_num_qubits)):
            tmp_res = scores_filtered_sorted_accordingly[i]
            max_score = max(tmp_res)
            for j in range(len(tmp_res)):
                plt.plot(i, tmp_res[j] / max_score, "b.", alpha=1.0, markersize=1.7, color=color_all)

            plt.plot(
                i,
                tmp_res[y_pred_sorted_accordingly[i]] / max_score,
                color=color_pred,
                marker=".",
                linestyle="None",
            )

        plt.xticks(
            list(range(0, len(scores_filtered), 30)),
            [qubit_list_sorted[i] for i in range(0, len(scores_filtered), 30)],
            fontsize=18,
        )
        plt.yticks(fontsize=18)
        plt.xlabel("Unseen test circuits (sorted along the number of qubits)", fontsize=18)
        plt.ylabel(
            "Evaluation scores of combinations of options \n (normalized per test circuit)",
            fontsize=18,
        )
        plt.tight_layout()

        plt.ylim(0, 1.05)
        plt.xlim(0, len(scores_filtered))
        result_path = Path("results")
        if not result_path.is_dir():
            result_path.mkdir()
        plt.savefig(result_path / "y_pred_eval_normed.pdf", bbox_inches="tight")

    def predict_probs(self, qasm_str_or_path: str | QuantumCircuit, figure_of_merit: reward.figure_of_merit) -> int:
        """Returns the probabilities for all supported quantum devices to be the most suitable one for the given quantum circuit.

        Args:
            qasm_str_or_path (str | QuantumCircuit): The qasm string or path to the qasm file
            figure_of_merit (reward.reward_functions): The figure of merit to be used for prediction

        Returns:
            int: The index of the predicted compilation option
        """

        if self.clf is None:
            path = ml.helper.get_path_trained_model(figure_of_merit)
            if path.is_file():
                self.clf = load(str(path))
            else:
                error_msg = "Classifier is neither trained nor saved."
                raise FileNotFoundError(error_msg)

        feature_dict = ml.helper.create_feature_dict(qasm_str_or_path)
        feature_vector = list(feature_dict.values())

        path = ml.helper.get_path_trained_model(figure_of_merit, return_non_zero_indices=True)
        non_zero_indices = np.load(path, allow_pickle=True)
        feature_vector = [feature_vector[i] for i in non_zero_indices]

        return cast(int, self.clf.predict_proba([feature_vector])[0])  # type: ignore[attr-defined]



from __future__ import annotations

import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any

import numpy as np
import requests
from mqt.bench.qiskit_helper import get_native_gates
from mqt.bench.utils import calc_supermarq_features, get_cmap_from_devicename
from mqt.predictor import reward, rl
from packaging import version
from pytket.architecture import Architecture
from pytket.circuit import Circuit, Node, Qubit
from pytket.passes import (
    CliffordSimp,
    FullPeepholeOptimise,
    PeepholeOptimise2Q,
    RemoveRedundancies,
    RoutingPass,
)
from pytket.placement import place_with_map
from qiskit import QuantumCircuit
from qiskit.circuit.equivalence_library import StandardEquivalenceLibrary
from qiskit.circuit.library import XGate, ZGate
from qiskit.transpiler import CouplingMap
from qiskit.transpiler.passes import (
    ApplyLayout,
    BasicSwap,
    BasisTranslator,
    Collect2qBlocks,
    CommutativeCancellation,
    CommutativeInverseCancellation,
    ConsolidateBlocks,
    CXCancellation,
    DenseLayout,
    Depth,
    EnlargeWithAncilla,
    FixedPoint,
    FullAncillaAllocation,
    GatesInBasis,
    InverseCancellation,
    MinimumPoint,
    Optimize1qGatesDecomposition,
    OptimizeCliffords,
    RemoveDiagonalGatesBeforeMeasure,
    SabreLayout,
    SabreSwap,
    Size,
    StochasticSwap,
    TrivialLayout,
    UnitarySynthesis,
)
from qiskit.transpiler.preset_passmanagers import common
from qiskit.transpiler.runningpassmanager import ConditionalController
from sb3_contrib import MaskablePPO
from tqdm import tqdm

if TYPE_CHECKING:
    from numpy.typing import NDArray

if TYPE_CHECKING or sys.version_info >= (3, 10, 0):
    from importlib import metadata, resources
else:
    import importlib_metadata as metadata
    import importlib_resources as resources

logger = logging.getLogger("mqt-predictor")


NUM_ACTIONS_OPT = 13
NUM_ACTIONS_LAYOUT = 3
NUM_ACTIONS_ROUTING = 4
NUM_ACTIONS_SYNTHESIS = 1
NUM_ACTIONS_TERMINATE = 1
NUM_ACTIONS_DEVICES = 7
NUM_ACTIONS_MAPPING = 1
NUM_FEATURE_VECTOR_ELEMENTS = 7


def qcompile(
    qc: QuantumCircuit | str,
    figure_of_merit: reward.figure_of_merit = "expected_fidelity",
    device_name: str = "ibm_washington",
    predictor_singleton: rl.Predictor | None = None,
) -> tuple[QuantumCircuit, list[str]]:
    """Compiles a given quantum circuit to a device optimizing for the given figure of merit.

    Args:
        qc (QuantumCircuit | str): The quantum circuit to be compiled. If a string is given, it is assumed to be a path to a qasm file.
        figure_of_merit (reward.reward_functions, optional): The figure of merit to be used for compilation. Defaults to "expected_fidelity".
        device_name (str, optional): The name of the device to compile to. Defaults to "ibm_washington".
        predictor_singleton (rl.Predictor, optional): A predictor object that is used for compilation. If None, a new predictor object is created. Defaults to None.

    Returns:
        tuple[QuantumCircuit, list[str]] | bool: Returns a tuple containing the compiled quantum circuit and the compilation information. If compilation fails, False is returned.
    """

    if predictor_singleton is None:
        predictor = rl.Predictor(figure_of_merit=figure_of_merit, device_name=device_name)
    else:
        predictor = predictor_singleton

    return predictor.compile_as_predicted(qc)


def get_actions_opt() -> list[dict[str, Any]]:
    """Returns a list of dictionaries containing information about the optimization passes that are available."""
    return [
        {
            "name": "Optimize1qGatesDecomposition",
            "transpile_pass": [Optimize1qGatesDecomposition()],
            "origin": "qiskit",
        },
        {
            "name": "CXCancellation",
            "transpile_pass": [CXCancellation()],
            "origin": "qiskit",
        },
        {
            "name": "CommutativeCancellation",
            "transpile_pass": [CommutativeCancellation()],
            "origin": "qiskit",
        },
        {
            "name": "CommutativeInverseCancellation",
            "transpile_pass": [CommutativeInverseCancellation()],
            "origin": "qiskit",
        },
        {
            "name": "RemoveDiagonalGatesBeforeMeasure",
            "transpile_pass": [RemoveDiagonalGatesBeforeMeasure()],
            "origin": "qiskit",
        },
        {
            "name": "InverseCancellation",
            "transpile_pass": [InverseCancellation([XGate(), ZGate()])],
            "origin": "qiskit",
        },
        {
            "name": "OptimizeCliffords",
            "transpile_pass": [OptimizeCliffords()],
            "origin": "qiskit",
        },
        {
            "name": "Opt2qBlocks",
            "transpile_pass": [Collect2qBlocks(), ConsolidateBlocks()],
            "origin": "qiskit",
        },
        {
            "name": "PeepholeOptimise2Q",
            "transpile_pass": [PeepholeOptimise2Q()],
            "origin": "tket",
        },
        {
            "name": "CliffordSimp",
            "transpile_pass": [CliffordSimp()],
            "origin": "tket",
        },
        {
            "name": "FullPeepholeOptimiseCX",
            "transpile_pass": [FullPeepholeOptimise()],
            "origin": "tket",
        },
        {
            "name": "RemoveRedundancies",
            "transpile_pass": [RemoveRedundancies()],
            "origin": "tket",
        },
        {
            "name": "QiskitO3",
            "transpile_pass": lambda bgates, cmap: [
                Collect2qBlocks(),
                ConsolidateBlocks(basis_gates=bgates),
                UnitarySynthesis(basis_gates=bgates, coupling_map=cmap),
                Optimize1qGatesDecomposition(basis=bgates),
                CommutativeCancellation(basis_gates=bgates),
                GatesInBasis(bgates),
                ConditionalController(
                    [
                        pass_
                        for x in common.generate_translation_passmanager(
                            target=None, basis_gates=bgates, coupling_map=cmap
                        ).passes()
                        for pass_ in x["passes"]
                    ],
                    condition=lambda property_set: not property_set["all_gates_in_basis"],
                ),
                Depth(recurse=True),
                FixedPoint("depth"),
                Size(recurse=True),
                FixedPoint("size"),
                MinimumPoint(["depth", "size"], "optimization_loop"),
            ],
            "origin": "qiskit",
            "do_while": lambda property_set: (not property_set["optimization_loop_minimum_point"]),
        },
    ]


def get_actions_layout() -> list[dict[str, Any]]:
    """Returns a list of dictionaries containing information about the layout passes that are available."""
    return [
        {
            "name": "TrivialLayout",
            "transpile_pass": lambda c: [
                TrivialLayout(coupling_map=CouplingMap(c)),
                FullAncillaAllocation(coupling_map=CouplingMap(c)),
                EnlargeWithAncilla(),
                ApplyLayout(),
            ],
            "origin": "qiskit",
        },
        {
            "name": "DenseLayout",
            "transpile_pass": lambda c: [
                DenseLayout(coupling_map=CouplingMap(c)),
                FullAncillaAllocation(coupling_map=CouplingMap(c)),
                EnlargeWithAncilla(),
                ApplyLayout(),
            ],
            "origin": "qiskit",
        },
        {
            "name": "SabreLayout",
            "transpile_pass": lambda c: [
                SabreLayout(coupling_map=CouplingMap(c), skip_routing=True),
                FullAncillaAllocation(coupling_map=CouplingMap(c)),
                EnlargeWithAncilla(),
                ApplyLayout(),
            ],
            "origin": "qiskit",
        },
    ]


def get_actions_routing() -> list[dict[str, Any]]:
    """Returns a list of dictionaries containing information about the routing passes that are available."""
    return [
        {
            "name": "BasicSwap",
            "transpile_pass": lambda c: [BasicSwap(coupling_map=CouplingMap(c))],
            "origin": "qiskit",
        },
        {
            "name": "RoutingPass",
            "transpile_pass": lambda c: [
                PreProcessTKETRoutingAfterQiskitLayout(),
                RoutingPass(Architecture(c)),
            ],
            "origin": "tket",
        },
        {
            "name": "StochasticSwap",
            "transpile_pass": lambda c: [StochasticSwap(coupling_map=CouplingMap(c))],
            "origin": "qiskit",
        },
        {
            "name": "SabreSwap",
            "transpile_pass": lambda c: [SabreSwap(coupling_map=CouplingMap(c))],
            "origin": "qiskit",
        },
    ]


def get_actions_mapping() -> list[dict[str, Any]]:
    """Returns a list of dictionaries containing information about the mapping passes that are available."""
    return [
        {
            "name": "SabreMapping",
            "transpile_pass": lambda c: [
                SabreLayout(coupling_map=CouplingMap(c), skip_routing=False),
            ],
            "origin": "qiskit",
        },
    ]


def get_actions_synthesis() -> list[dict[str, Any]]:
    """Returns a list of dictionaries containing information about the synthesis passes that are available."""
    return [
        {
            "name": "BasisTranslator",
            "transpile_pass": lambda g: [BasisTranslator(StandardEquivalenceLibrary, target_basis=g)],
            "origin": "qiskit",
        },
    ]


def get_action_terminate() -> dict[str, Any]:
    """Returns a dictionary containing information about the terminate pass that is available."""
    return {"name": "terminate"}


def get_devices() -> list[dict[str, Any]]:
    """Returns a list of dictionaries containing information about the devices that are available."""
    return [
        {
            "name": "ibm_washington",
            "cmap": get_cmap_from_devicename("ibm_washington"),
            "native_gates": get_native_gates("ibm"),
            "max_qubits": 127,
        },
        {
            "name": "ibm_montreal",
            "cmap": get_cmap_from_devicename("ibm_montreal"),
            "native_gates": get_native_gates("ibm"),
            "max_qubits": 27,
        },
        {
            "name": "oqc_lucy",
            "cmap": get_cmap_from_devicename("oqc_lucy"),
            "native_gates": get_native_gates("oqc"),
            "max_qubits": 8,
        },
        {
            "name": "rigetti_aspen_m2",
            "cmap": get_cmap_from_devicename("rigetti_aspen_m2"),
            "native_gates": get_native_gates("rigetti"),
            "max_qubits": 80,
        },
        {
            "name": "ionq_harmony",
            "cmap": get_cmap_from_devicename("ionq_harmony"),
            "native_gates": get_native_gates("ionq"),
            "max_qubits": 11,
        },
        {
            "name": "ionq_aria1",
            "cmap": get_cmap_from_devicename("ionq_aria1"),
            "native_gates": get_native_gates("ionq"),
            "max_qubits": 25,
        },
        {
            "name": "quantinuum_h2",
            "cmap": get_cmap_from_devicename("quantinuum_h2"),
            "native_gates": get_native_gates("quantinuum"),
            "max_qubits": 32,
        },
    ]


def get_state_sample(max_qubits: int | None = None) -> tuple[QuantumCircuit, str]:
    """Returns a random quantum circuit from the training circuits folder.

    Args:
        max_qubits (int, None): The maximum number of qubits the returned quantum circuit may have. If no limit is set, it defaults to None.

    Returns:
        tuple[QuantumCircuit, str]: A tuple containing the random quantum circuit and the path to the file from which it was read.
    """
    file_list = list(get_path_training_circuits().glob("*.qasm"))

    path_zip = get_path_training_circuits() / "training_data_compilation.zip"
    if len(file_list) == 0 and path_zip.exists():
        import zipfile

        with zipfile.ZipFile(str(path_zip), "r") as zip_ref:
            zip_ref.extractall(get_path_training_circuits())

        file_list = list(get_path_training_circuits().glob("*.qasm"))
        assert len(file_list) > 0

    found_suitable_qc = False
    while not found_suitable_qc:
        random_index = np.random.randint(len(file_list))
        num_qubits = int(str(file_list[random_index]).split("_")[-1].split(".")[0])
        if max_qubits and num_qubits > max_qubits:
            continue
        found_suitable_qc = True

    try:
        qc = QuantumCircuit.from_qasm_file(str(file_list[random_index]))
    except Exception:
        raise RuntimeError("Could not read QuantumCircuit from: " + str(file_list[random_index])) from None

    return qc, str(file_list[random_index])


def create_feature_dict(qc: QuantumCircuit) -> dict[str, int | NDArray[np.float_]]:
    """Creates a feature dictionary for a given quantum circuit.

    Args:
        qc (QuantumCircuit): The quantum circuit for which the feature dictionary is created.

    Returns:
        dict[str, Any]: The feature dictionary for the given quantum circuit.
    """

    feature_dict = {
        "num_qubits": qc.num_qubits,
        "depth": qc.depth(),
    }

    supermarq_features = calc_supermarq_features(qc)
    # for all dict values, put them in a list each
    feature_dict["program_communication"] = np.array([supermarq_features.program_communication], dtype=np.float32)
    feature_dict["critical_depth"] = np.array([supermarq_features.critical_depth], dtype=np.float32)
    feature_dict["entanglement_ratio"] = np.array([supermarq_features.entanglement_ratio], dtype=np.float32)
    feature_dict["parallelism"] = np.array([supermarq_features.parallelism], dtype=np.float32)
    feature_dict["liveness"] = np.array([supermarq_features.liveness], dtype=np.float32)

    return feature_dict


def get_path_training_data() -> Path:
    """Returns the path to the training data folder used for RL training."""
    return Path(str(resources.files("mqt.predictor"))) / "rl" / "training_data"


def get_path_trained_model() -> Path:
    """Returns the path to the trained model folder used for RL training."""
    return get_path_training_data() / "trained_model"


def get_path_training_circuits() -> Path:
    """Returns the path to the training circuits folder used for RL training."""
    return get_path_training_data() / "training_circuits"


def load_model(model_name: str) -> MaskablePPO:
    """Loads a trained model from the trained model folder.

    Args:
        model_name (str): The name of the model to be loaded.

    Returns:
        MaskablePPO: The loaded model.
    """
    path = get_path_trained_model()

    if Path(path / (model_name + ".zip")).exists():
        return MaskablePPO.load(path / (model_name + ".zip"))
    logger.info("Model does not exist. Try to retrieve suitable Model from GitHub...")
    try:
        mqtpredictor_module_version = metadata.version("mqt.predictor")
    except ModuleNotFoundError:
        error_msg = (
            "Could not retrieve version of mqt.predictor. Please run 'pip install . or pip install mqt.predictor'."
        )
        raise RuntimeError(error_msg) from None

    version_found = False
    response = requests.get("https://api.github.com/repos/cda-tum/mqt-predictor/tags")
    available_versions = []
    if not response:
        error_msg = "Querying the GitHub API failed. One reasons could be that the limit of 60 API calls per hour and IP address is exceeded."
        raise RuntimeError(error_msg)
    for elem in response.json():
        available_versions.append(elem["name"])
    for possible_version in available_versions:
        if version.parse(mqtpredictor_module_version) >= version.parse(possible_version):
            url = "https://api.github.com/repos/cda-tum/mqt-predictor/releases/tags/" + possible_version
            response = requests.get(url)
            if not response:
                error_msg = "Suitable trained models cannot be downloaded since the GitHub API failed. One reasons could be that the limit of 60 API calls per hour and IP address is exceeded."
                raise RuntimeError(error_msg)

            response_json = response.json()
            if "assets" in response_json:
                assets = response_json["assets"]
            elif "asset" in response_json:
                assets = [response_json["asset"]]
            else:
                assets = []

            for asset in assets:
                if model_name in asset["name"]:
                    version_found = True
                    download_url = asset["browser_download_url"]
                    logger.info("Downloading model from: " + download_url)
                    handle_downloading_model(download_url, model_name)
                    break

        if version_found:
            break

    if not version_found:
        error_msg = "No suitable model found on GitHub. Please update your mqt.predictor package using 'pip install -U mqt.predictor'."
        raise RuntimeError(error_msg) from None

    return MaskablePPO.load(path / model_name)


def handle_downloading_model(download_url: str, model_name: str) -> None:
    """Downloads a trained model from the given URL and saves it to the trained model folder.

    Args:
        download_url (str): The URL from which the model is downloaded.
        model_name (str): The name of the model to be downloaded.
    """
    logger.info("Start downloading model...")

    r = requests.get(download_url)
    total_length = int(r.headers.get("content-length"))  # type: ignore[arg-type]
    fname = str(get_path_trained_model() / (model_name + ".zip"))

    with Path(fname).open(mode="wb") as f, tqdm(
        desc=fname,
        total=total_length,
        unit="iB",
        unit_scale=True,
        unit_divisor=1024,
    ) as bar:
        for data in r.iter_content(chunk_size=1024):
            size = f.write(data)
            bar.update(size)
    logger.info(f"Download completed to {fname}. ")


class PreProcessTKETRoutingAfterQiskitLayout:
    """
    Pre-processing step to route a circuit with tket after a Qiskit Layout pass has been applied.
    The reason why we can apply the trivial layout here is that the circuit is already mapped by qiskit to the
    device qubits and its qubits are sorted by their ascending physical qubit indices.
    The trivial layout indices that this layout of the physical qubits is the identity mapping.
    """

    def apply(self, circuit: Circuit) -> None:
        """Applies the pre-processing step to route a circuit with tket after a Qiskit Layout pass has been applied."""
        mapping = {Qubit(i): Node(i) for i in range(circuit.n_qubits)}
        place_with_map(circuit=circuit, qmap=mapping)


def get_device(device_name: str) -> dict[str, Any]:
    """Returns the device with the given name.

    Args:
        device_name (str): The name of the device to be returned.

    Returns:
        dict[str, Any]: The device with the given name.
    """
    devices = get_devices()
    for device in devices:
        if device["name"] == device_name:
            return device

    msg = "No suitable device found."
    raise RuntimeError(msg)


def get_device_index_of_device(device_name: str) -> int:
    """Returns the index of the device with the given name.

    Args:
        device_name (str): The name of the device to be returned.

    Returns:
        int: The index of the device with the given name.
    """
    devices = get_devices()
    for i, device in enumerate(devices):
        if device["name"] == device_name:
            return i

    msg = "No suitable device found."
    raise RuntimeError(msg)



from pathlib import Path

from mqt.bench import get_benchmark
from mqt.predictor import rl
from mqt.predictor.evaluation import evaluate_sample_circuit


def test_evaluate_sample_circuit() -> None:
    qc = get_benchmark("ghz", 1, 3)
    filename = "test_3.qasm"
    qc.qasm(filename=filename)
    res = evaluate_sample_circuit(filename)
    expected_keys = []
    for compilation_setup in ["qiskit", "tket", "mqt-predictor_expected_fidelity", "mqt-predictor_critical_depth"]:
        for key in ["time", "expected_fidelity", "critical_depth"]:
            if "mqt-predictor" in compilation_setup:
                expected_keys.append(compilation_setup + "_" + key)
            else:
                for device in rl.helper.get_devices():
                    expected_keys.append(compilation_setup + "_" + device["name"] + "_" + key)

    assert all(key in res for key in expected_keys)
    if Path(filename).exists():
        Path(filename).unlink()



from mqt.predictor.rl.helper import qcompile
from mqt.predictor.rl.Predictor import Predictor
from mqt.predictor.rl.PredictorEnv import PredictorEnv

__all__ = [
    "qcompile",
    "Predictor",
    "PredictorEnv",
]



from pathlib import Path

import pytest
from mqt.bench import get_benchmark
from mqt.predictor import rl


def test_get_actions_opt() -> None:
    assert len(rl.helper.get_actions_opt()) == rl.helper.NUM_ACTIONS_OPT


def test_get_actions_layout() -> None:
    assert len(rl.helper.get_actions_layout()) == rl.helper.NUM_ACTIONS_LAYOUT


def test_et_actions_routing() -> None:
    assert len(rl.helper.get_actions_routing()) == rl.helper.NUM_ACTIONS_ROUTING


def test_get_actions_synthesis() -> None:
    assert len(rl.helper.get_actions_synthesis()) == rl.helper.NUM_ACTIONS_SYNTHESIS


def test_get_action_terminate() -> None:
    assert len(rl.helper.get_action_terminate()) == rl.helper.NUM_ACTIONS_TERMINATE


def test_get_actions_devices() -> None:
    assert len(rl.helper.get_devices()) == rl.helper.NUM_ACTIONS_DEVICES


def test_get_actions_mapping() -> None:
    assert len(rl.helper.get_actions_mapping()) == rl.helper.NUM_ACTIONS_MAPPING


@pytest.mark.parametrize(
    "device",
    ["ibm_washington", "ibm_montreal", "rigetti_aspen_m2", "oqc_lucy", "ionq_harmony", "ionq_aria1", "quantinuum_h2"],
)
def test_get_device(device: str) -> None:
    assert rl.helper.get_device(device)


def test_create_feature_dict() -> None:
    qc = get_benchmark("dj", 1, 5)
    features = rl.helper.create_feature_dict(qc)
    assert features
    assert len(features) == rl.helper.NUM_FEATURE_VECTOR_ELEMENTS


def test_get_path_trained_model() -> None:
    path = rl.helper.get_path_trained_model()
    assert path.exists()
    assert isinstance(path, Path)


def test_get_path_training_circuits() -> None:
    path = rl.helper.get_path_training_circuits()
    assert path.exists()
    assert isinstance(path, Path)
